//
// Keyboard routines for the HP series 300 HIL Keyboard
// (c) Anders Gustafsson, anders.gustafsson@pedago.fi
//
//
#include "system.h"
#include <libpic30.h>        // __delayXXX() functions macros defined here
#include <stdio.h>

#include "hil_kbd.h"
#include "ps2_mouse.h"
//
volatile unsigned int Counter;
volatile int TXCounter;
volatile unsigned int TXRegister;
volatile unsigned int RXCounter;
volatile unsigned int RXRegister;
volatile unsigned char Key = 0;
volatile unsigned char Mouse = 0;
int Ticker = 0;
int Language =0x1F; // US English
//
char HILAddress = 0;

void HardReset(void);

typedef union {
    struct {
        unsigned configbits              :1; 
        unsigned status1                 :1;
        unsigned status2                 :1;
        unsigned status3                 :1;
        unsigned status4                 :1;
        unsigned status5                 :1;
        unsigned shift                 :1;
        unsigned ctrl                 :1;
    };
    unsigned char status;
} KbdStatus_t;

KbdStatus_t KbdStatus;
char Loopback = 1;
// Circular buffer (FiFo) for HIL reception 
//
#define BUFFER_SIZE 128
static volatile struct {
	int		ri, wi, ct;			/* Read index, Write index, Data counter */
	int	buff[BUFFER_SIZE];      /* FIFO buffer, received serial */
} RxFifo;

static volatile struct {
	int		ri, wi, ct;			/* Read index, Write index, Data counter */
	int	buff[BUFFER_SIZE];      /* FIFO buffer, received serial */
} TxFifo;
//
// HIL Command mnemonics for debugging
char *HILNames[] = {"IFC", "EPT", "ELB", "IDD", "DSR", "PST","RRG","WRG",
                        "ACF0","ACF1","ACF2","ACF3","ACF4","ACF5","ACF6","ACF7",
                        //"ACF8","ACF9","ACFa","ACFb","ACFc","ACFd","ACFe","ACFf",
                        "POL0","POL1","POL2","POL3","POL4","POL5","POL5","POL7",
                        "POL8","POL9","POLa","POLb","POLc","POLd","POLe","POLf",
                        "RPL0","RPL1","RPL2","RPL3","RPL4","RPL5","RPL6","RPL7",
                        "RPL8","RPL9","RPLa","RPLb","RPLc","RPLd","RPLe","RPLf",
                        "RNM","RST", "EXD","RSC","034","035","036","037",
                        "038","039","03a","03b","03c","DKA","EK1","EK2",
                        "PR1","PR2","PR3","PR4","PR5","PR6","PR7", "PRM",                // 0x40
                        "AK1","AK2","AK3","AK4","AK5","AK6","AK6", "ACK",
                        "050","051","052","053","054","055","056","057",                //0x50
                        "058","059","05a","05b","05c","05d","05e","05f",
                        "060","061","062","063","064","065","066","067",                //0x60
                        "068","069","06a","06b","06c","06d","06e","06f",
                        "070","071","072","073","074","075","076","077",                //0x70
                        "078","079","07a","07b","07c","07d","07e","07f",
                        "080","081","082","083","084","085","086","057",
                        "088","089","08a","08b","08c","08d","08e","08f",                //0x80 - 0xEF is device specific
                        "090","091","092","093","094","095","096","087",
                        "098","099","09a","09b","09c","0ad","09e","09f",
                        "0a0","0a1","0a2","0a3","0a4","0a5","0a6","0a7",
                        "0a8","0a9","0aa","0ab","0ac","0bd","0ae","0af",
                        "0b0","0b1","0b2","0b3","0b4","0b5","0b6","0b7",
                        "0b8","0b9","0ba","0bb","0bc","0bd","0be","0bf",
                        "0c0","0c1","0c2","0c3","0c4","0c5","0c6","0c7",
                        "0c8","0c9","0ca","0cb","0cc","0cd","0ce","0cf",
                        "0d0","0d1","0d2","0d3","0d4","0d5","0d6","0d7",
                        "0d8","0d9","0da","0db","0dc","0dd","0de","0df",
                        "0e0","0e1","0e2","0e3","0e4","0e5","0e6","0e7",
                        "0e8","0e9","0ea","0eb","0ec","0ed","0ee","0ef",
                        "0F0","0F1","0F2","0F3","0F4","0F5","0F6","0f7",                //0xf0 .. 0xf9 reserved
                        "0F8","0F9","RIO","SHR","TER","CAE","DHR","0ff"
};
extern int x;
extern int y;
extern int z;
extern int addr;
extern int cmd;
extern int data;
extern int pollno;
//
//
// Init HIL keyboard interface. HIL has no clock, just data in and out
//
void HILInit(void)
{
    KbdStatus.status = 0;
    //
    AD1PCFG = 0xFFFF;           // Turn off analogue
    TRISBbits.TRISB15 = 0;      // Diag leds
    TRISBbits.TRISB14 = 0;
    TRISBbits.TRISB13 = 1;
    //
    // Pins:
    // B7 - Data line input, also INT0
    // B5 - Data line output
    //
    TRISBbits.TRISB5 = 0;       // B5 is signal output
    TRISBbits.TRISB7 = 1;       // B7 Signal Input for INT0
    //TRISBbits.TRISB8 = 0;       // B8 is trigger output
    //PORTBbits.RB8 = 1;
    ODCBbits.ODB8 = 0;
    //ODCBbits.ODB8 = 1;          // Make it open drain.
    //
    // https://engineering.purdue.edu/ece477/Archive/2009/Spring/S09-Grp06/Code/PIC/pic24_code_examples/docs/int0__bounce_8c-source.html
    //
    // Set up for INT0, Start bit detect
    //
    _INT0IF = 0;                //Clear the interrupt flag
    _INT0IP = 7;                //Choose a priority
    _INT0EP = 1;                //negative edge triggerred
    _INT0IE = 1;                //enable INT0 interrupt
    //
    // Testing 115200 baud. Bit time 8,68us
    // Timer clock Fosc/2 = 32MHz, 278 ticks
    //
    OC1CON1 = 0;
    OC1CON2 = 0;
    OC1CON1bits.OCTSEL = 7;     // System clock
    OC1CON1bits.OCM = 0b011;
    //
    OC1CON2bits.SYNCSEL = 0b11111;
    OC1CON2bits.OCTRIS = 1;
    //
    OC1RS = 160; //138;
    OC1R = 0;
    //
    // Output compare 2 for reception
    //
    IFS0bits.OC2IF = 0;     // Enable Output Compare 1 interrupt
    IEC0bits.OC2IE = 1;
    IPC1bits.OC2IP = 7;
    //
    OC2CON1 = 0;
    OC2CON2 = 0;
    OC2CON1bits.OCTSEL = 7; // System clock
    OC2CON1bits.OCM = 0;    //0b101; //0b011;
    //
    OC2CON2bits.SYNCSEL = 0b11111;
    OC2CON2bits.OCTRIS = 1;
    //
    OC2RS = 160; //138;
    OC2R = 0;
    //
    IFS0bits.OC2IF = 0;     // Disable Output Compare 2 interrupt
    IEC0bits.OC2IE = 0;
    IPC1bits.OC2IP = 7;
    //
    // UART
    //
    _U1RXIE = 0;
	_U1TXIE = 0;
    //unlock registers
    __builtin_write_OSCCONL(OSCCON & 0xBF);
    // Configure Output Functions (Table 9-2)
    // Assign U1TX To Pin RP21 (C5)
    RPOR10bits.RP21R = 3;
    TRISCbits.TRISC5 = 0; // output
	// Initialize UART1
    U1BRG = FCY / 2 / 4 / 115200 - 1;
    _BRGH = 1;
	_UARTEN = 1;
	_UTXEN = 1;
    //
    printf("\r\nHPHILKBD %s Starting. %s %s", VERSION, __DATE__, __TIME__);
    //
    // Read dipswitch
    //
    TRISB = TRISB | 0b00001111; // Lower 4 bits of language code 0x00..0x1F
    TRISCbits.TRISC0 = 1;       // Last bit thereof
    Language = PORTB & 0x0F;
    if(PORTCbits.RC0) Language = Language + 0x10;
    printf("\r\nPORTB: %X", PORTB);
    printf("\r\nLanguage code is %X", Language);        
}
//
// Parse and act on HIL-command
//
void HILParse(bool cmd, unsigned char address, unsigned char data)
{
    int bytes = 0;
    switch(data)
    {
        case 0x00:                          // Interface clear
            //MARK
            HILSend(cmd, address, data);
            break;
        case 0x01:                          // EPT Enter passthrough
                                            // If we emulate two devices then
                                            // EPT on the second stops echo
            //printf(" %X - %X ", address, HILAddress);
            if(address == HILAddress+1)
            {
                //printf(" Match ");
                Loopback = 0;
            }
            if(Loopback==1) HILSend(cmd, address, data);
            //printf(" loopback %d", Loopback);
            break;  
        case 0x02:                          // ELB Enter loopback
            //MARK                            // If we emulate two devices then
            //printf("\r\n %X ", address);  // we answer in the second
            if(address == HILAddress+1) Loopback = 1;
            if(Loopback==1) HILSend(cmd, address, data);
            //printf(" loopback2 %d", Loopback);
            break;
        case 0x03:                         // Device ID

            if(address == HILAddress && Loopback==1)
            {
                HILSend(0, HILAddress, 0xC0 + Language); 
                //HILSend(0, HILAddress, 0xDF);      // ITF keyboard 0xC0, US 0x1F
                //HILSend(0, HILAddress, 0x1F);      // Vectra, US
                //HILSend(0, HILAddress, 0xCE);      // ITF keyboard, Swe
                HILSend(0, HILAddress, 0x00);      // 
                //HILSend(cmd, HILAddress, data);
                printf(" send device ID %X ", 0xDF);
            }
            if(address == 2 && Loopback==1)
            {

                HILSend(0, 2, 0x68);      // Mouse
                HILSend(0, 2, 0x12);      // 
                HILSend(0, 2, 0xc2);      //
                HILSend(0, 2, 0x1e);
                HILSend(0, 2, 0x0f);
                //HILSend(cmd, HILAddress, data);
                printf(" send device ID %X ", 0x68);
            }
            if(Loopback) HILSend(cmd, address, data);
            break;
        case 0x05:                          // Selftest
            if(Loopback==1)
            {
            HILSend(0, HILAddress, 0x00);      // Success
            HILSend(cmd, HILAddress, data);    // Retransmit cmd
            }
            break;
        case 0x08:                          // Auto configure
        case 0x09:
        case 0x0a:
        case 0x0b:
        case 0x0c:
        case 0x0d:
        case 0x0e:
        case 0x0f:
            //__delay_us(100);


            HILAddress = data & 0b111;      // Three last bits is my address
            if(Loopback==1) HILSend(cmd, address, data+2);
            //printf(" My address %d, loopback is %d rtn %d", HILAddress, Loopback, data);
            break;
        case 0x10:                          // Poll
            //printf(" Poll %d", Loopback);
            ++Ticker;
            if(TX_test() && Loopback==1)// && Ticker > 5)
            {
                Key = TX_getc();
                Ticker=0;

                //if(Ticker % 2) Key = 0xda; //0x52;
                //else Key = 0xdb; //0x53;
                HILSend(0, HILAddress, 0x40);      // Record header
                HILSend(0, HILAddress, Key);      // Character data/scancode (1 dn)
                HILSend(cmd, address, data+2);    // Retransmit cmd, incremented by
                                            // number of bytes sent
                printf(" H: %X ", Key);
                if(Key == 0x0E) Key = 0x0F;
                else Key = 0;
            }
            if(Mouse && !TX_test() && Loopback)
            {
                //
                // Check if any of the mouse buttons has changed
                //
                if(MouseStatus.right != OldMouseStatus.right || MouseStatus.left != OldMouseStatus.left || MouseStatus.middle != OldMouseStatus.middle)
                    HILSend(0, HILAddress+1, 0x42);         // Header, X and Y, scancode
                else
                    HILSend(0, HILAddress+1, 0x02);         // just x and y
                HILSend(0, HILAddress+1, mousex); 
                HILSend(0, HILAddress+1, mousey);
                bytes = 3;
                //
                // Buttons?
                //
                if(MouseStatus.right && !OldMouseStatus.right)
                {
                   HILSend(0, HILAddress+1, 0x82); // Button 1 down
                   ++bytes;
                   printf("\r\nB 1 dn");
                }
                if(!MouseStatus.right && OldMouseStatus.right)
                {
                   HILSend(0, HILAddress+1, 0x83); // Button 1 up
                   ++bytes;
                   printf("\r\nB 1 up");
                }
                if(MouseStatus.left && !OldMouseStatus.left)
                {
                   HILSend(0, HILAddress+1, 0x80); // Button 2 down
                   ++bytes;
                   printf("\r\nB 2 dn");
                }
                if(!MouseStatus.left && OldMouseStatus.left)
                {
                   HILSend(0, HILAddress+1, 0x81); // Button 1 up
                   ++bytes;
                   printf("\r\nB 2 up");
                }
                // Middle button
                if(MouseStatus.middle && !OldMouseStatus.middle)
                {
                   HILSend(0, HILAddress+1, 0x84); // Button 3 (middle) down
                   ++bytes;
                   printf("\r\nB 3 dn");
                }
                if(!MouseStatus.middle && OldMouseStatus.middle)
                {
                   HILSend(0, HILAddress+1, 0x85); // Button  up
                   ++bytes;
                   printf("\r\nB 3 up");
                }
                HILSend(cmd, address, data+bytes);    // Retransmit cmd, incremented
                Mouse = 0;

                OldMouseStatus.status = MouseStatus.status;
                //printf("  Mouse poll: %d %d \r\n", mousex,mousey);
                mousex = 0;
                mousey = 0;
            }
            if(!Key && !Mouse && Loopback==1) HILSend(cmd, address, data);
            break;
        case 0xfe: 
                printf(" Hard reset");
                Loopback = 1;
                Key = 0;
                HardReset();
                break;
        default:
            break;
    }
}
//
// Check if received chars
//
int RX_test (void)
{
	return RxFifo.ct;           // Returns number of bytes in the Rx FIFO
}
//
// Read one int from buffer (blocking))
//
int RX_getc (void)
{
	int d;
	int i;

	while (!RxFifo.ct) ;		// Wait while Rx FIFO empty

	i = RxFifo.ri;				// Get a byte from Rx FIFO
	d = RxFifo.buff[i++];
	RxFifo.ri = i % BUFFER_SIZE;
    __builtin_disable_interrupts();
	RxFifo.ct--;
     __builtin_enable_interrupts();
	return d;
}
//
// Send data over HIL. Stuff data into TXRegister. Initialise counter
// and turn on timer interrupt.
//
// Parity:
// unsigned char parity = 0;
// while(data)
// {
// parity^=(data&1);
// data >>1;
// }
unsigned char ReverseByte (unsigned char In)
{
    unsigned char Out=0;
    int i;
    for(i=0;i<8;i++)
    {
        if(In &(0x01 <<i))
            Out|=(0x080>>i);
    }
    return Out;
}

unsigned char ReverseAddress (unsigned char In)
{
    unsigned char Out=0;
    int i;
    for(i=0;i<4;i++)
    {
        if(In &(0x01 <<i))
            Out|=(0x04>>i);
    }
    return Out;
}
//
// Send data or command as HIL-packet
//
int HILSend(bool cmd, char address, char data)
{
    //__delay_us(150);
    int HILData = 0;
    int temp = 0;
    if(cmd) HILData = HILData | 0b0000000000001000;
    HILData = HILData | ReverseAddress(address);
    temp = ReverseByte(data);
    temp = temp << 4;
    HILData = HILData | temp;
    HILPutch(HILData);
    //if(data!=0x10) printf("\r\nSent %X %X %X", cmd, address, data);
    return 1;
}
//
// Send one integer over HIL, add parity if needed
//
int HILPutch(int data)
{
    int i=15;
    int parity=0;
    int tempdata;
    //
    while(IEC0bits.OC1IE)   // Wait until last char sent
        ;

    TXRegister = data<<1;   // Start bit
    TXRegister = TXRegister | 0b1100000000000000;   // Stop bits
    tempdata=TXRegister;
    //
    while(i--)
    {
        parity^=(tempdata&1);   // Parity will be 0 if even number of ones
        tempdata = tempdata >>1;
    }
    if(!parity) TXRegister = TXRegister | 0b010000000000000;
    
    TXCounter = 16;
    while(OC1TMR)           // Wait until start
        ;

    IEC0bits.OC1IE = 1;     // Enable interrupts
    return 0;
}
//
// Put a byte into the TX buffer for transmission
// returns number of bytes in buffer
int HILPutTX(char data)
{
    int i;
    //printf(" Put: %X ", data);
    i = TxFifo.ct;				// Number of bytes in the FIFO
    if (i < BUFFER_SIZE) {		// Skip if FIFO is full
        TxFifo.ct = ++i;
        i = TxFifo.wi;
        TxFifo.buff[i++] = data;          // Store received data into the FIFO
        TxFifo.wi = i % BUFFER_SIZE;	 // Next write ptr
    }
    return i;
}

//
// Check if buffer has characters to send
//
int TX_test (void)
{
	return TxFifo.ct;           // Returns number of bytes in the Tx FIFO
}
//
// Read one byte to send from TX buffer from buffer (blocking))
// Make sure to test i there is something in the buffer before calling!
int TX_getc (void)
{
	int d;
	int i;

	while (!TxFifo.ct) ;		// Wait while Tx FIFO empty

	i = TxFifo.ri;				// Get a byte from Tx FIFO
	d = TxFifo.buff[i++];
	TxFifo.ri = i % BUFFER_SIZE;
    //__builtin_disable_interrupts();
	TxFifo.ct--;
    // __builtin_enable_interrupts();
    //printf(" Get: %X ", d&0xff);
	return d;
}
//
// Interrupt routines
//
//
// Data interrupt. Start bit. Turn off this interrupt and start bit-
// counter.
//
void __attribute__((__interrupt__, auto_psv)) _INT0Interrupt( void )
{
    RXCounter = 16;
    RXRegister = 0;
    OC2TMR = 0;             // Start OC2 counter in known state
    OC2CON1bits.OCM = 0b011;
    IEC0bits.OC2IE = 1;     // enable OC2 interrupt
    _INT0IE = 0;            // disable INT0 interrupt
    _INT0IF = 0;            // clear the interrupt bit
}
//
//
// Output compare 1 interrupt. Transmit timer
// Shift out the bits in TXRegister and turn off timer interrupt when done.
//
void __attribute__((__interrupt__, auto_psv)) _OC1Interrupt( void )
{

    if(TXCounter <= 0)      // All done? We arrive here one last time
    {
        LATBbits.LATB5 = 1;
        IEC0bits.OC1IE = 0;  
    }
    else LATBbits.LATB5 = TXRegister & 1;
    TXRegister = TXRegister >> 1;
    TXCounter--;

    IFS0bits.OC1IF = 0;     // clear the interrupt bit
}
//
// Output compare 2 interrupt. Receive timer. Count in bits
// and put in fifo when all received
//
void __attribute__((__interrupt__, auto_psv)) _OC2Interrupt( void )
{
    int i;
   
    if(PORTBbits.RB7) RXRegister = RXRegister | 0b1000000000000000;
    //
    RXCounter--;
    if(RXCounter == 0)              // All done?
    {
        IEC0bits.OC2IE = 0;
        OC2CON1bits.OCM = 0;
        i = RxFifo.ct;				// Number of bytes in the FIFO
        if (i < BUFFER_SIZE) {		// Skip if FIFO is full
            RxFifo.ct = ++i;
            i = RxFifo.wi;
            RxFifo.buff[i++] = RXRegister;          // Store received data into the FIFO
            RxFifo.wi = i % BUFFER_SIZE;	 // Next write ptr
        }
        _INT0IE = 1;                //enable INT0 interrupt for next rx
    }
    else RXRegister = RXRegister >> 1;
    IFS0bits.OC2IF = 0;             // clear the interrupt bit
    _INT0IF = 0;                    //clear the interrupt bit
}
