//
// PS/2 Keyboard routines for the HP series 300 HIL Keyboard emulator
// (c) Anders Gustafsson, anders.gustafsson@pedago.fi
//
//
#include <xc.h>
#include "system.h"

#include <libpic30.h>        // __delayXXX() functions macros defined here
#include <stdio.h>

#include "ps2_kbd.h"

//
volatile unsigned int PS2RXCounter;
volatile unsigned int PS2RXRegister;
volatile unsigned char PS2Key = 0;


typedef union {
    struct {
        unsigned configbits              :1; 
        unsigned status1                 :1;
        unsigned status2                 :1;
        unsigned status3                 :1;
        unsigned status4                 :1;
        unsigned status5                 :1;
        unsigned shift                 :1;
        unsigned ctrl                 :1;
    };
    unsigned char status;
} PS2KbdStatus_t;

PS2KbdStatus_t PS2KbdStatus;
// Circular buffer (FiFo) for reception 
//
#define BUFFER_SIZE 128
static volatile struct {
	int		ri, wi, ct;			/* Read index, Write index, Data counter */
	int	buff[BUFFER_SIZE];      /* FIFO buffer, received serial */
} PS2Fifo;
//

//
// Use this with the Microchip MLA USB host sample
//
// First you need to call void PS2Init(void). Then check for incoming
// by calling PS2Test, then get it with PS2getc
//
//
// Check if received chars
//
int PS2Test (void)
{
	return PS2Fifo.ct;           // Returns number of bytes in the Rx FIFO
}
//
// Read one int from buffer (blocking))
//
int PS2getc (void)
{
	int d;
	int i;

	while (!PS2Fifo.ct) ;		// Wait while Rx FIFO empty

	i = PS2Fifo.ri;				// Get a byte from Rx FIFO
	d = PS2Fifo.buff[i++];
	PS2Fifo.ri = i % BUFFER_SIZE;
    __builtin_disable_interrupts();
	PS2Fifo.ct--;
     __builtin_enable_interrupts();
     //
     //printf("\r\n**%X ",d&0xff);
	return d&0xff;         // Mask startbit and parity/stop
}
//
// Init keyboard interface. 
//
void PS2Init(void)
{

    PS2KbdStatus.status = 0;
    //
    // Read Config, language RB15-13
    //
    AD1PCFG = 0xFFFF;       // Turn off analogue
    //
    //
    // Pins:
    // C7 - Data line input, 5V tolerant!
    // C6 - Clock line input, INT1, 5V tolerant!
    //
    TRISCbits.TRISC7 = 1;       // C7 is signal input
    TRISCbits.TRISC6 = 1;       // C6 is clock, INT1
    //
    // https://engineering.purdue.edu/ece477/Archive/2009/Spring/S09-Grp06/Code/PIC/pic24_code_examples/docs/int0__bounce_8c-source.html
    //
    // Set up for INT1, Clock sense/counter
    //
    RPINR0bits.INT1R = 22;       // INT 1 is RP22
    _INT1IF = 0;                // Clear the interrupt flag
    _INT1IP = 6;                // Choose a priority
    _INT1EP = 0;                // positive edge triggerred
    _INT1IE = 1;                // enable INT1 interrupt
    //
    PS2RXCounter = 11;          // Start + 8 data + parity + stop;
    PS2RXRegister = 0;
    printf("\r\nHPHILKBD-PS2 Starting. %s %s", __DATE__, __TIME__);
//    while(1)
//    {
//        if(PS2Test())
//            printf("\r\n %X", PS2getc());
//    }

}
//
// Reads 0..9  a..f and enter and returns 0..F, enter is returned as 0x20
//char PS2Scan2Hex (char ch)
//{
//    uint8_t index2;
//    for(index2 = 0; index2 < sizeof(PS2keyTranslationTable2)/sizeof(PS2_HP_TRANSLATION_TABLE_ENTRY); index2++)
//    {
//        //printf("\r\n%d %X %X ", index2, ch,PS2keyTranslationTable2[index2].key);
//        if(PS2keyTranslationTable2[index2].key == ch)
//        {
//            return PS2keyTranslationTable2[index2].scancode;
//        }
//    }
//    return 0;
//}
//
// Interrupt routines
//
//
// Clock interrupt. Sample bit on every rising edge
//
void __attribute__((__interrupt__, auto_psv)) _INT1Interrupt( void )
{
    int i;
    if(PORTCbits.RC7) PS2RXRegister = PS2RXRegister | 0b0000000100000000;
    //
    PS2RXCounter--;
    if(PS2RXCounter == 0)           // All done?
    {
        i = PS2Fifo.ct;				// Number of bytes in the FIFO
        if (i < BUFFER_SIZE) {		// Skip if FIFO is full
            PS2Fifo.ct = ++i;
            i = PS2Fifo.wi;
            PS2Fifo.buff[i++] = PS2RXRegister; // Store received data into the FIFO
            PS2Fifo.wi = i % BUFFER_SIZE;	 // Next write ptr
        }
        PS2RXCounter = 11;         // Init counter for next
    }
    else PS2RXRegister = PS2RXRegister >> 1;

    _INT1IF = 0;                    // clear the interrupt bit
}
//
// Reset input counter (not needed)
//
void PS2Reset(void)
{
    //PS2RXCounter = 11;          // Start + 8 data + parity + stop;
    //PS2RXRegister = 0;
}