//
// PS/2 Mouse routines for the HP series 300 HIL Keyboard emulator
// (c) Anders Gustafsson, anders.gustafsson@pedago.fi
//
//
#include <xc.h>
#include "system.h"

#include <libpic30.h>        // __delayXXX() functions macros defined here
#include <stdio.h>

#include "ps2_mouse.h"

//
volatile unsigned int PS2MouseRXCounter;
volatile unsigned int PS2MouseRXRegister;
volatile unsigned int PS2MouseTXCounter;
volatile unsigned int PS2MouseTXRegister;
volatile unsigned int PS2MouseTX;
//
MouseStatus_t MouseStatus;
MouseStatus_t OldMouseStatus;
//
int MouseErrors;

// Circular buffer (FiFo) for reception 
//
#define BUFFER_SIZE 128
static volatile struct {
	int		ri, wi, ct;			/* Read index, Write index, Data counter */
	int	buff[BUFFER_SIZE];      /* FIFO buffer, received serial */
} PS2MouseFifo;
//
//
// First you need to call void PS2MouseInit(void). Then check for incoming
// by calling 
//
//
// Check if received chars
//
int PS2MouseTest (void)
{
	return PS2MouseFifo.ct;           // Returns number of bytes in the Rx FIFO
}
//
// Read one int from buffer (blocking))
//
int PS2Mousegetc (void)
{
	int d;
	int i;

	while (!PS2MouseFifo.ct) ;		// Wait while Rx FIFO empty

	i = PS2MouseFifo.ri;				// Get a byte from Rx FIFO
	d = PS2MouseFifo.buff[i++];
	PS2MouseFifo.ri = i % BUFFER_SIZE;
    __builtin_disable_interrupts();
	PS2MouseFifo.ct--;
     __builtin_enable_interrupts();
    //
	return d&0xff;         // Mask startbit and parity/stop
}
//
// Send one to mouse
//
void PS2Mouseputc(char ch)
{
    int i=9;        // Parity calculated over the data bits
    int tempdata;
    int parity=0;
    
    PS2MouseTXRegister = ch<<1;   // Start bit
    PS2MouseTXRegister = PS2MouseTXRegister & 0b1111110111111111;   // Stop bits
    tempdata=ch;
    //
    while(i--)
    {
        parity^=(tempdata&1);   // Parity will be 0 if even number of ones
        tempdata = tempdata >>1;
    }
    if(parity) PS2MouseTXRegister = PS2MouseTXRegister | 0b000001000000000;
    
    
    PS2MouseTXCounter = 11;          // Start + 8 data + parity
    
    _INT2IE = 0;                    // disable INT2 interrupt
    PS2MouseTX = 1;
    TRISCbits.TRISC9 = 0;           // C9 is now output
    TRISCbits.TRISC8 = 0;           // C8 is now output
    //
    LATCbits.LATC8 = 0;             // Make clock low
    __delay_us(120);                // Wait more than 100us
    LATCbits.LATC9 = 0;             // Make data low
    __delay_us(10);
    _INT2IE = 1;                    // enable INT2 interrupt
    TRISCbits.TRISC8 = 1;           // Make clock high/input. Mouse should now 
                                    // start sending clock pulses
}
//
// Init mouse interface. 
//
void PS2MouseInit(void)
{
    int i = 2;
    int timeout = 100;
    //
    AD1PCFG = 0xFFFF;       // Turn off analogue
    //
    //
    // Pins:
    // C9 - Data line input, 5V tolerant!
    // C8 - Clock line input, INT2, 5V tolerant!
    //
    TRISCbits.TRISC9 = 1;       // C9 is signal input
    TRISCbits.TRISC8 = 1;       // C8 is clock, INT2
    TRISCbits.TRISC3 = 0;       // Mouse power
    ODCCbits.ODC8 = 1;          // Open drain on both
    ODCCbits.ODC9 = 1;
	//
    // Set up for INT2, Clock sense/counter
    //
    RPINR1bits.INT2R = 24;       // INT 2 is RP24
    _INT2IF = 0;                // Clear the interrupt flag
    _INT2IP = 6;                // Choose a priority
    _INT2EP = 0;                // positive edge triggerred
    //
    //
    PS2MouseRXCounter = 11;          // Start + 8 data + parity + stop;
    PS2MouseRXRegister = 0;
    PS2MouseTXCounter = 11;          // Start + 8 data + parity + stop;
    PS2MouseTXRegister = 0;
    PS2MouseTX = 0;
    MouseErrors = 0;
    //
    // Power up mouse
    //
    LATCbits.LATC3 = 1;
    __delay_ms(100);
    //
    _INT2IF = 0; 
    _INT2IE = 1;                // enable INT2 interrupt
    printf("\r\nHPHILKBD-PS2 Mouse Starting. %s %s", __DATE__, __TIME__);
    // wait for initial id
    while(i && timeout !=0)
    {
        if(PS2MouseTest())
        {
            printf("\r\n%d %X", i, PS2Mousegetc());
            i--;
        }
        __delay_ms(10);
        --timeout;
    }
    //
    // Nothing from mouse? Assume nonexistent
    //
    if(timeout == 0)
    {
        printf("\r\nNo mouse present");
        return;
    }
    PS2Mouseputc(0xFF);             // Reset mouse, then wait for 2 bytes
    i = 3;
    timeout = 1000;
    while(i && timeout !=0)
    {
        if(PS2MouseTest())
        {
            printf("\r\n %X", PS2Mousegetc());
            i--;
        }
        __delay_ms(10);
        --timeout;
    }

    PS2Mouseputc(0xF4);             // Set reporting mode
    timeout = 1000;
    while(!PS2MouseTest() && timeout !=0)          // ..and wait for the reply
    {
        --timeout;
        __delay_ms(10);
    }
    if(PS2MouseTest()) 
    PS2Mousegetc();
}
//
// Reset counters i case of error
//
void PS2MouseReset(void)
{
    PS2MouseRXCounter = 11;          // Start + 8 data + parity + stop;
    PS2MouseRXRegister = 0;
    PS2MouseTXCounter = 11;          // Start + 8 data + parity + stop;
    PS2MouseTXRegister = 0;
    PS2MouseTX = 0;
    MouseErrors = 0;   
}
//
// Interrupt routines
//
//
// Clock interrupt. Sample bit on every rising edge on reception
// send bits on every falling
//
void __attribute__((__interrupt__, auto_psv)) _INT2Interrupt( void )
{
    int i;
    
    if(PS2MouseTX==0)
    {
        MARK
        if(PORTCbits.RC9) PS2MouseRXRegister = PS2MouseRXRegister | 0b0000001000000000;
        //
        PS2MouseRXCounter--;
        if(PS2MouseRXCounter == 0)           // All done?
        {
            i = PS2MouseFifo.ct;				// Number of bytes in the FIFO
            if (i < BUFFER_SIZE) {              // Skip if FIFO is full
                PS2MouseFifo.ct = ++i;
                i = PS2MouseFifo.wi;
                PS2MouseFifo.buff[i++] = PS2MouseRXRegister; // Store received data into the FIFO
                PS2MouseFifo.wi = i % BUFFER_SIZE;	 // Next write ptr
            }
            PS2MouseRXCounter = 11;         // Init counter for next
            PS2MouseRXRegister = 0;
        }
        else PS2MouseRXRegister = PS2MouseRXRegister >> 1;
    }
    else
    {
        if(PS2MouseTXCounter == 0)      // All done?
        {
            _INT2IE = 0; 
            
            LATCbits.LATC9 = PS2MouseTXRegister & 1;
            while(PORTCbits.RC8)        // Wait for clock to rise
                ;
            Nop();                      // Give us some margin
            Nop();
            TRISCbits.TRISC9 = 1;       // Put back as inputs
            TRISCbits.TRISC8 = 1;       //
            while(PORTCbits.RC9)        // Wait for data to rise
                ;
            PS2MouseRXCounter = 11;     // Clear out possible rx garbage
            PS2MouseRXRegister = 0;
            PS2MouseTX = 0;
            _INT2IE = 1; 

        }
        else
        {
            LATCbits.LATC9 = PS2MouseTXRegister & 1;
            //MARK
        }
        PS2MouseTXRegister = PS2MouseTXRegister >> 1;
        PS2MouseTXCounter--;

    }
    _INT2IF = 0;                    // clear the interrupt bit
}
